#pragma once
#include <petscdmplex.h>
#include <petscsnes.h>
#include <petscds.h>

//typedef enum {NEUMANN, DIRICHLET} BCType;
//typedef enum {RUN_FULL, RUN_TEST_MS} RunType;

// --------------
// Stokes context
// --------------

typedef struct {
    PetscReal     viscosity;
} Stokes_AppCtx;

// ---------------
// General context
// ---------------

typedef struct {
    PetscBool     test_ms;           /* Whether to run tests with MS, or solve the full problem */
    PetscLogEvent createMeshEvent;
    PetscBool     showInitial, showSolution, showError, computeL2error;
    PetscBool     saveVTK;
    char          saveSolutionVTK[PETSC_MAX_PATH_LEN];
    char          gmsh_file[PETSC_MAX_PATH_LEN];
    /* Domain and mesh definition */
    PetscInt      dim;               /* The topological mesh dimension */
    PetscBool     interpolate;       /* Generate intermediate mesh elements */
    PetscBool     simplex;           /* Use simplices or tensor product cells */
    PetscReal     refinementLimit;   /* The largest allowable cell volume */
    PetscBool     testPartition;     /* Use a fixed partitioning for testing */
    PetscBool     setNullSpace;
    /* Problem definition */
    char          bndLabelName[PETSC_MAX_PATH_LEN];
    /* Homogeneus dirichlet boundary */
    PetscInt*     hom_dirichlet_id;
    PetscInt      num_hom_dirichlet_id;
    /* Non-homogeneous dirichlet boundary */
    PetscInt*     nonhom_dirichlet_id;
    PetscInt      num_nonhom_dirichlet_id;
    PetscReal*    nonhom_dirichlet_values;
    /* Neumann boundaries */
    PetscInt*     neumann_id;
    PetscInt      num_neumann_id;
    PetscReal*    neumann_values;
    //BCType        bcType;
    PetscErrorCode (**bndFuncs)(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *u, void *ctx);
    PetscErrorCode (**exactFuncs)(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *u, void *ctx);

    Stokes_AppCtx stk_ctx;
} AppCtx;

// ---------------
// Problem context
// ---------------

typedef struct {
    PetscErrorCode (*CreateMesh)(MPI_Comm comm, AppCtx *user, DM *dm);
    PetscErrorCode (*SetupProblem)(DM dm, AppCtx *user);
    PetscErrorCode (*SetupDiscretization)(DM dm, AppCtx *user);
} GeneralProblem;
