#pragma once
#include <petscdmplex.h>
#include <petscsnes.h>
#include <petscds.h>

#include "appctx.h"

#include "petscdmplex.h"
#include "petscsnes.h"
#include "petscds.h"

#include "appctx.h"
#include "meshes.h"
#include "aux.h"

PetscErrorCode stokes_test_u_exact_2d(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *u, void *ctx)
{
    u[0] = x[0]*x[0] + x[1]*x[1];
    u[1] = 2.0*x[0]*x[0] - 2.0*x[0]*x[1];
    return 0;
}

PetscErrorCode stokes_test_p_exact_2d(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *p, void *ctx)
{
    *p = x[0] + x[1] - 1.0;
    return 0;
}

PetscErrorCode stokes_test_u_exact_3d(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *u, void *ctx)
{
    u[0] = x[0]*x[0] + x[1]*x[1];
    u[1] = x[1]*x[1] + x[2]*x[2];
    u[2] = x[0]*x[0] + x[1]*x[1] - 2.0*(x[0] + x[1])*x[2];
    return 0;
}

PetscErrorCode stokes_test_p_exact_3d(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *p, void *ctx)
{
    *p = x[0] + x[1] + x[2] - 1.5;
    return 0;
}

void stokes_test_f0_u(PetscInt dim, PetscInt Nf, PetscInt NfAux,
                 const PetscInt uOff[], const PetscInt uOff_x[], const PetscScalar u[], const PetscScalar u_t[], const PetscScalar u_x[],
                 const PetscInt aOff[], const PetscInt aOff_x[], const PetscScalar a[], const PetscScalar a_t[], const PetscScalar a_x[],
                 PetscReal t, const PetscReal x[], PetscScalar f0[])
{
    for (PetscInt c = 0; c < dim; ++c) f0[c] = 3.0;
}

PetscErrorCode stokes_viscosity(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *a, void *ctx)
{
    AppCtx* app_ctx = (AppCtx*) ctx;
    *a = app_ctx->stk_ctx.viscosity;
    return 0;
}

PetscErrorCode stokes_wall_u(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *u, void *ctx)
{
    for(PetscInt c=0; c<dim; ++c){ u[c] = 0.0; }
    return 0;
}

PetscErrorCode stokes_constant_u(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *u, void *ctx)
{
    AppCtx* app_ctx = (AppCtx*) ctx;
    for(PetscInt c=0; c<dim; ++c){ u[c] = app_ctx->nonhom_dirichlet_values[c]; }
    return 0;
}

void stokes_bnd_p(PetscInt dim, PetscInt Nf, PetscInt NfAux,
                       const PetscInt uOff[], const PetscInt uOff_x[], const PetscScalar u[], const PetscScalar u_t[], const PetscScalar u_x[],
                       const PetscInt aOff[], const PetscInt aOff_x[], const PetscScalar a[], const PetscScalar a_t[], const PetscScalar a_x[],
                       PetscReal t, const PetscReal x[], const PetscReal n[], PetscScalar f0[])
{
    f0[0]=0.0;
}

void stokes_pres_inlet(PetscInt dim, PetscInt Nf, PetscInt NfAux,
                const PetscInt uOff[], const PetscInt uOff_x[], const PetscScalar u[], const PetscScalar u_t[], const PetscScalar u_x[],
                const PetscInt aOff[], const PetscInt aOff_x[], const PetscScalar a[], const PetscScalar a_t[], const PetscScalar a_x[],
                PetscReal t, const PetscReal x[], const PetscReal n[], PetscScalar f0[])
{
    PetscReal p_inlet = 0.008; // should be specified as auxiliary field.
    for(PetscInt c=0; c<dim; ++c){ f0[c] = -p_inlet*n[c]; }
}

void stokes_pres_outlet(PetscInt dim, PetscInt Nf, PetscInt NfAux,
                       const PetscInt uOff[], const PetscInt uOff_x[], const PetscScalar u[], const PetscScalar u_t[], const PetscScalar u_x[],
                       const PetscInt aOff[], const PetscInt aOff_x[], const PetscScalar a[], const PetscScalar a_t[], const PetscScalar a_x[],
                       PetscReal t, const PetscReal x[], const PetscReal n[], PetscScalar f0[])
{
    PetscReal p_outlet = 0.1; // should be specified as auxiliary field.
    for(PetscInt c=0; c<dim; ++c){ f0[c] = -p_outlet*n[c]; }
}

void stokes_f0_u(PetscInt dim, PetscInt Nf, PetscInt NfAux,
          const PetscInt uOff[], const PetscInt uOff_x[], const PetscScalar u[], const PetscScalar u_t[], const PetscScalar u_x[],
          const PetscInt aOff[], const PetscInt aOff_x[], const PetscScalar a[], const PetscScalar a_t[], const PetscScalar a_x[],
          PetscReal t, const PetscReal x[], PetscScalar f0[])
{
  for (PetscInt c = 0; c < dim; ++c) f0[c] = 0.0;
}

/* gradU[comp*dim+d] = {u_x, u_y, v_x, v_y} or {u_x, u_y, u_z, v_x, v_y, v_z, w_x, w_y, w_z}
   u[Ncomp]          = {p} */
void stokes_f1_u(PetscInt dim, PetscInt Nf, PetscInt NfAux,
          const PetscInt uOff[], const PetscInt uOff_x[], const PetscScalar u[], const PetscScalar u_t[], const PetscScalar u_x[],
          const PetscInt aOff[], const PetscInt aOff_x[], const PetscScalar a[], const PetscScalar a_t[], const PetscScalar a_x[],
          PetscReal t, const PetscReal x[], PetscScalar f1[])
{
  const PetscInt Ncomp = dim;
  const PetscReal  nu  = a[0];
  PetscInt       comp, d;

  for (comp = 0; comp < Ncomp; ++comp) {
    for (d = 0; d < dim; ++d) {
      f1[comp*dim+d] = nu*u_x[comp*dim+d];
      //f1[comp*dim+d] = 0.5*nu*(u_x[comp*dim+d] + u_x[d*dim+comp]);
      //f1[comp*dim+d] = u_x[comp*dim+d];
    }
    f1[comp*dim+comp] -= u[Ncomp];
  }
}

// gradU[comp*dim+d] = {u_x, u_y, v_x, v_y} or {u_x, u_y, u_z, v_x, v_y, v_z, w_x, w_y, w_z}
void stokes_f0_p(PetscInt dim, PetscInt Nf, PetscInt NfAux,
          const PetscInt uOff[], const PetscInt uOff_x[], const PetscScalar u[], const PetscScalar u_t[], const PetscScalar u_x[],
          const PetscInt aOff[], const PetscInt aOff_x[], const PetscScalar a[], const PetscScalar a_t[], const PetscScalar a_x[],
          PetscReal t, const PetscReal x[], PetscScalar f0[])
{
  PetscInt d;
  for (d = 0, f0[0] = 0.0; d < dim; ++d) f0[0] += u_x[d*dim+d];
  // f0[0] = u_x[0] + u_x[4] + u_x[8] = u_x + v_y + w_z.
}

void stokes_f1_p(PetscInt dim, PetscInt Nf, PetscInt NfAux,
          const PetscInt uOff[], const PetscInt uOff_x[], const PetscScalar u[], const PetscScalar u_t[], const PetscScalar u_x[],
          const PetscInt aOff[], const PetscInt aOff_x[], const PetscScalar a[], const PetscScalar a_t[], const PetscScalar a_x[],
          PetscReal t, const PetscReal x[], PetscScalar f1[])
{
  PetscInt d;
  for (d = 0; d < dim; ++d) f1[d] = 0.0;
}

/* < q, \nabla\cdot u >
   NcompI = 1, NcompJ = dim */
void stokes_g1_pu(PetscInt dim, PetscInt Nf, PetscInt NfAux,
           const PetscInt uOff[], const PetscInt uOff_x[], const PetscScalar u[], const PetscScalar u_t[], const PetscScalar u_x[],
           const PetscInt aOff[], const PetscInt aOff_x[], const PetscScalar a[], const PetscScalar a_t[], const PetscScalar a_x[],
           PetscReal t, PetscReal u_tShift, const PetscReal x[], PetscScalar g1[])
{
  PetscInt d;
  for (d = 0; d < dim; ++d) g1[d*dim+d] = 1.0;
}

/* -< \nabla\cdot v, p >
    NcompI = dim, NcompJ = 1 */
void stokes_g2_up(PetscInt dim, PetscInt Nf, PetscInt NfAux,
           const PetscInt uOff[], const PetscInt uOff_x[], const PetscScalar u[], const PetscScalar u_t[], const PetscScalar u_x[],
           const PetscInt aOff[], const PetscInt aOff_x[], const PetscScalar a[], const PetscScalar a_t[], const PetscScalar a_x[],
           PetscReal t, PetscReal u_tShift, const PetscReal x[], PetscScalar g2[])
{
  PetscInt d;
  for (d = 0; d < dim; ++d) g2[d*dim+d] = -1.0; /* \frac{\partial\psi^{u_d}}{\partial x_d} */
}

void stokes_g3_uu(PetscInt dim, PetscInt Nf, PetscInt NfAux,
           const PetscInt uOff[], const PetscInt uOff_x[], const PetscScalar u[], const PetscScalar u_t[], const PetscScalar u_x[],
           const PetscInt aOff[], const PetscInt aOff_x[], const PetscScalar a[], const PetscScalar a_t[], const PetscScalar a_x[],
           PetscReal t, PetscReal u_tShift, const PetscReal x[], PetscScalar g3[])
{
  const PetscInt Ncomp = dim;
  const PetscReal   nu = a[0];

  //< \nabla v, \nabla u >
  for (PetscInt fc=0; fc<Ncomp; ++fc) {
    for (PetscInt d=0; d<dim; ++d) {
      g3[((fc*Ncomp+fc)*dim+d)*dim+d] = nu;
    }
  }
  //< \nabla v, {\nabla u}^T >: IT IS WRONG!
  /*
  for(PetscInt fc=0; fc<Ncomp; ++fc){
      d=fc;
      for(PetscInt gc=0; gc<Ncomp; ++gc){
          PetscInt d2=gc;
          g3[((fc*Ncomp+gc)*dim+d)*dim+d2] += 0.5*nu;
      }
  }*/
}

#undef __FUNCT__
#define __FUNCT__ "Stokes_ProcessOptions"
PetscErrorCode Stokes_ProcessOptions(MPI_Comm comm, AppCtx *options)
{
    PetscErrorCode ierr;
    options->stk_ctx.viscosity = 1.0;

    PetscFunctionBeginUser;
    ierr = PetscOptionsBegin(comm, "", "Stokes Problem Options", "DMPLEX");CHKERRQ(ierr);
    ierr = PetscOptionsReal("-stokes_viscosity", "Viscosity of stokes problem", "fmcm.c", options->stk_ctx.viscosity, &options->stk_ctx.viscosity, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsEnd();

    PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "Stokes_Setup"
PetscErrorCode Stokes_Setup(DM dm, AppCtx *user)
{
  PetscDS        prob;
  PetscErrorCode ierr;

  PetscFunctionBeginUser;
  ierr = DMGetDS(dm, &prob);CHKERRQ(ierr);
  ierr = PetscDSSetResidual(prob, 0, user->test_ms? stokes_test_f0_u : NULL, stokes_f1_u);CHKERRQ(ierr);
  ierr = PetscDSSetResidual(prob, 1, stokes_f0_p, stokes_f1_p);CHKERRQ(ierr);
  ierr = PetscDSSetJacobian(prob, 0, 0, NULL, NULL,  NULL,  stokes_g3_uu);CHKERRQ(ierr);
  ierr = PetscDSSetJacobian(prob, 0, 1, NULL, NULL,  stokes_g2_up, NULL);CHKERRQ(ierr);
  ierr = PetscDSSetJacobian(prob, 1, 0, NULL, stokes_g1_pu, NULL,  NULL);CHKERRQ(ierr);
  if(user->num_neumann_id != 0) {
    ierr = PetscDSSetBdResidual(prob, 0, stokes_pres_inlet, NULL); CHKERRQ(ierr);
    ierr = PetscDSSetBdResidual(prob, 1, stokes_bnd_p, NULL); CHKERRQ(ierr);
  }
  /*if(user->stk_ctx.num_pres_outlet_id != 0) {
    ierr = PetscDSSetBdResidual(prob, 0, stokes_pres_outlet, NULL); CHKERRQ(ierr);
    ierr = PetscDSSetBdResidual(prob, 1, stokes_bnd_p, NULL); CHKERRQ(ierr);
  }*/
  if(user->test_ms) {
      user->bndFuncs[0] = user->dim==2? stokes_test_u_exact_2d : stokes_test_u_exact_3d;
      user->bndFuncs[1] = user->dim==2? stokes_test_p_exact_2d : stokes_test_p_exact_3d;
      user->exactFuncs[0] = user->dim==2? stokes_test_u_exact_2d : stokes_test_u_exact_3d;
      user->exactFuncs[1] = user->dim==2? stokes_test_p_exact_2d : stokes_test_p_exact_3d;
      user->computeL2error = PETSC_TRUE;
      user->setNullSpace = PETSC_TRUE;
  }
  else {
      user->bndFuncs[0] = stokes_wall_u;
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "Stokes_Auxiliar"
static PetscErrorCode Stokes_Auxiliar(DM dm, DM dmAux, AppCtx *user)
{
    PetscErrorCode (*matFuncs[1])(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar u[], void *ctx) = {stokes_viscosity};
    Vec            nu;
    PetscErrorCode ierr;
    AppCtx **urs = {&user};
    
    PetscFunctionBegin;
    ierr = DMCreateLocalVector(dmAux, &nu);CHKERRQ(ierr);
    ierr = DMProjectFunction(dmAux, 0.0, matFuncs, (void**)urs, INSERT_ALL_VALUES, nu);CHKERRQ(ierr);
    ierr = PetscObjectCompose((PetscObject) dm, "A", (PetscObject) nu);CHKERRQ(ierr);
    ierr = VecDestroy(&nu);CHKERRQ(ierr);
    PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "Stokes_SetupDiscretization"
PetscErrorCode Stokes_SetupDiscretization(DM dm, AppCtx *user)
{
    DM              cdm   = dm, dmAux, coordDM;
    const PetscInt  dim   = user->dim;
    //const PetscInt  id    = 29;
    PetscFE         fe[2], feAux, feBnd[2];
    PetscQuadrature q;
    PetscDS         prob, probAux;
    PetscInt        order;
    PetscErrorCode  ierr;
    
    PetscFunctionBeginUser;
    ierr = PetscMalloc(2 * sizeof(void (*)(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *u, void *ctx)), &user->bndFuncs);CHKERRQ(ierr);
    ierr = PetscMalloc(2 * sizeof(void (*)(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *u, void *ctx)), &user->exactFuncs);CHKERRQ(ierr);
    ierr = Stokes_ProcessOptions(PETSC_COMM_WORLD, user); CHKERRQ(ierr);
    /* Create finite element */
    ierr = PetscFECreateDefault(dm, dim, dim, user->simplex, "vel_", -1, &fe[0]);CHKERRQ(ierr);
    ierr = PetscObjectSetName((PetscObject) fe[0], "velocity");CHKERRQ(ierr);
    ierr = PetscFEGetQuadrature(fe[0], &q);CHKERRQ(ierr);
    ierr = PetscQuadratureGetOrder(q, &order);CHKERRQ(ierr);
    ierr = PetscFECreateDefault(dm, dim, 1, user->simplex, "pres_", order, &fe[1]);CHKERRQ(ierr);
    ierr = PetscObjectSetName((PetscObject) fe[1], "pressure");CHKERRQ(ierr);
    // viscosity as auxiliary field
    ierr = PetscFECreateDefault(dm, dim, 1, user->simplex, "viscosity_", -1, &feAux);CHKERRQ(ierr);
    ierr = PetscObjectSetName((PetscObject) feAux, "viscosity");CHKERRQ(ierr);
    ierr = PetscFESetQuadrature(feAux, q);CHKERRQ(ierr);
    // boundaries.
    if(user->num_neumann_id != 0) {
      ierr = PetscFECreateDefault(dm, dim-1, dim, user->simplex, "velBnd_", order, &feBnd[0]);CHKERRQ(ierr);
      ierr = PetscObjectSetName((PetscObject) feBnd[0], "velBnd_");CHKERRQ(ierr);
      ierr = PetscFECreateDefault(dm, dim-1, 1, user->simplex, "presBnd_", order, &feBnd[1]);CHKERRQ(ierr);
      ierr = PetscObjectSetName((PetscObject) feBnd[1], "presBnd_");CHKERRQ(ierr);
    }
    /* Set discretization and boundary conditions for each mesh */
    while (cdm) {
        ierr = DMGetDS(cdm, &prob);CHKERRQ(ierr);
        ierr = PetscDSSetDiscretization(prob, 0, (PetscObject) fe[0]);CHKERRQ(ierr);
        ierr = PetscDSSetDiscretization(prob, 1, (PetscObject) fe[1]);CHKERRQ(ierr);
        if(user->num_neumann_id != 0) {
            ierr = PetscDSSetBdDiscretization(prob, 0, (PetscObject) feBnd[0]); CHKERRQ(ierr);
            ierr = PetscDSSetBdDiscretization(prob, 1, (PetscObject) feBnd[1]); CHKERRQ(ierr);
        }
        ierr = Stokes_Setup(cdm, user);CHKERRQ(ierr);

        ierr = DMGetCoordinateDM(cdm, &coordDM); CHKERRQ(ierr);
        ierr = DMClone(cdm, &dmAux);CHKERRQ(ierr);
        ierr = DMSetCoordinateDM(dmAux, coordDM);CHKERRQ(ierr);
        ierr = DMSetApplicationContext(dmAux, user); CHKERRQ(ierr);
        ierr = DMGetDS(dmAux, &probAux);CHKERRQ(ierr);
        ierr = PetscDSSetDiscretization(probAux, 0, (PetscObject) feAux);CHKERRQ(ierr);
        ierr = PetscObjectCompose((PetscObject) cdm, "dmAux", (PetscObject) dmAux);CHKERRQ(ierr);
        ierr = Stokes_Auxiliar(cdm, dmAux, user);CHKERRQ(ierr);
        ierr = DMDestroy(&dmAux);CHKERRQ(ierr);

        if(user->num_hom_dirichlet_id != 0) {
            ierr = DMAddBoundary(cdm, PETSC_TRUE, "dirichlet", user->bndLabelName, 0, 0, NULL, (void (*)()) user->bndFuncs[0],
                                 user->num_hom_dirichlet_id, user->hom_dirichlet_id, user); CHKERRQ(ierr);
        }
        /*if(user->stk_ctx.num_vel_id != 0) {
            ierr = DMAddBoundary(cdm, PETSC_TRUE, "vel", "Face Sets", 0, 0, NULL, (void (*)()) stokes_constant_u,
                                 user->stk_ctx.num_vel_id, user->stk_ctx.vel_id, user); CHKERRQ(ierr);
        }*/
        if(user->num_neumann_id != 0) { // to be concluded after tests..
            ierr = DMAddBoundary(cdm, PETSC_FALSE, "neumann", user->bndLabelName, 0, 0, NULL, NULL,
                                 user->num_neumann_id, user->neumann_id, user); CHKERRQ(ierr);
        }
        /*if(user->stk_ctx.num_pres_outlet_id != 0) { // to be concluded after tests..
            ierr = DMAddBoundary(cdm, PETSC_FALSE, "pres_outlet", "Face Sets", 0, 0, NULL, NULL,
                                 user->stk_ctx.num_pres_outlet_id, user->stk_ctx.pres_outlet_id, user); CHKERRQ(ierr);
        }*/
        
        ierr = DMGetCoarseDM(cdm, &cdm);CHKERRQ(ierr);
    }
    ierr = PetscFEDestroy(&fe[0]);CHKERRQ(ierr);
    ierr = PetscFEDestroy(&fe[1]);CHKERRQ(ierr);
    if(user->num_neumann_id != 0) {
        ierr = PetscFEDestroy(&feBnd[0]);CHKERRQ(ierr);
        ierr = PetscFEDestroy(&feBnd[1]);CHKERRQ(ierr);
    }
    ierr = PetscFEDestroy(&feAux);CHKERRQ(ierr);
    PetscFunctionReturn(0);
}
